@@ -63,6 +63,9 @@ public:
     Fraction maxFramerate() const;
     void setMaxFramerate(const Fraction &framerate);
 
+    int maxPendingFrames() const;
+    void setMaxPendingFrames(int newMaxBufferSize);
+
     virtual int64_t framePts(const std::optional<std::chrono::nanoseconds> &presentationTimestamp)
     {
         return std::chrono::duration_cast<std::chrono::milliseconds>(presentationTimestamp.value()).count();
@@ -78,7 +81,6 @@ public:
     }
 
     void stateChanged(pw_stream_state state);
-    friend class PipeWireProduceThread;
     void setupStream();
     virtual void processFrame(const PipeWireFrame &frame);
     void render(const QImage &image, const PipeWireFrame &frame);
@@ -91,6 +93,8 @@ public:
 
     void setQuality(const std::optional<quint8> &quality);
 
+    void setEncodingPreference(const PipeWireBaseEncodedStream::EncodingPreference &encodingPreference);
+
     const uint m_nodeId;
     QScopedPointer<PipeWireSourceStream> m_stream;
     QString m_error;
@@ -104,6 +108,8 @@ public:
 
     std::optional<quint8> m_quality;
 
+    PipeWireBaseEncodedStream::EncodingPreference m_encodingPreference;
+
     struct {
         QImage texture;
         std::optional<QPoint> position;
@@ -128,6 +134,11 @@ public:
     std::atomic_int m_pendingFilterFrames = 0;
     std::atomic_int m_pendingEncodeFrames = 0;
 
+    // Controls how many frames we can push into ffmpeg's encoding stream
+    std::atomic_int m_maxPendingFrames = 50;
+
+    Fraction m_maxFramerate = {60, 1};
+
 Q_SIGNALS:
     void producedFrames();
 
