@@ -9,6 +9,7 @@
 #include "glhelpers.h"
 #include "logging.h"
 #include "pipewiresourcestream.h"
+#include "pwhelpers.h"
 
 #include <QGuiApplication>
 #include <QOpenGLContext>
@@ -38,12 +39,13 @@ public:
     uint m_nodeId = 0;
     std::optional<uint> m_fd;
     std::function<QSGTexture *()> m_createNextTexture;
-    QScopedPointer<PipeWireSourceStream> m_stream;
-    QScopedPointer<QOpenGLTexture> m_texture;
+    std::unique_ptr<PipeWireSourceStream> m_stream;
+    std::unique_ptr<QOpenGLTexture> m_texture;
 
     EGLImage m_image = nullptr;
     bool m_needsRecreateTexture = false;
     bool m_allowDmaBuf = true;
+    bool m_ready = false;
 
     struct {
         QImage texture;
@@ -82,7 +84,6 @@ PipeWireSourceItem::PipeWireSourceItem(QQuickItem *parent)
     , d(new PipeWireSourceItemPrivate)
 {
     setFlag(ItemHasContents, true);
-    setEnabled(false);
 }
 
 PipeWireSourceItem::~PipeWireSourceItem()
@@ -97,7 +98,7 @@ void PipeWireSourceItem::itemChange(QQuickItem::ItemChange change, const QQuickI
     switch (change) {
     case ItemVisibleHasChanged:
         if (!isVisible()) {
-            setEnabled(false);
+            setReady(false);
         }
         if (d->m_stream) {
             d->m_stream->setActive(isVisible());
@@ -117,7 +118,7 @@ void PipeWireSourceItem::itemChange(QQuickItem::ItemChange change, const QQuickI
 void PipeWireSourceItem::releaseResources()
 {
     if (window() && (d->m_image || d->m_texture)) {
-        window()->scheduleRenderJob(new DiscardEglPixmapRunnable(d->m_image, d->m_texture.take()), QQuickWindow::NoStage);
+        window()->scheduleRenderJob(new DiscardEglPixmapRunnable(d->m_image, d->m_texture.release()), QQuickWindow::NoStage);
         d->m_image = EGL_NO_IMAGE_KHR;
     }
 }
@@ -141,7 +142,7 @@ void PipeWireSourceItem::resetFd()
         return;
     }
 
-    setEnabled(false);
+    setReady(false);
     close(*d->m_fd);
     d->m_fd.reset();
     d->m_stream.reset(nullptr);
@@ -153,7 +154,7 @@ void PipeWireSourceItem::resetFd()
 
 void PipeWireSourceItem::refresh()
 {
-    setEnabled(false);
+    setReady(false);
 
     if (!isComponentComplete()) {
         return;
@@ -182,8 +183,8 @@ void PipeWireSourceItem::refresh()
         }
         d->m_stream->setActive(isVisible());
 
-        connect(d->m_stream.data(), &PipeWireSourceStream::frameReceived, this, &PipeWireSourceItem::processFrame);
-        connect(d->m_stream.data(), &PipeWireSourceStream::stateChanged, this, &PipeWireSourceItem::stateChanged);
+        connect(d->m_stream.get(), &PipeWireSourceStream::frameReceived, this, &PipeWireSourceItem::processFrame);
+        connect(d->m_stream.get(), &PipeWireSourceStream::stateChanged, this, &PipeWireSourceItem::stateChanged);
     }
     Q_EMIT stateChanged();
 }
@@ -263,18 +264,14 @@ QSGNode *PipeWireSourceItem::updatePaintNode(QSGNode *node, QQuickItem::UpdatePa
         return nullptr;
     }
 
-    QSGImageNode *screenNode;
-    auto pwNode = dynamic_cast<PipeWireRenderNode *>(node);
+    auto pwNode = static_cast<PipeWireRenderNode *>(node);
     if (!pwNode) {
-        delete node;
         pwNode = new PipeWireRenderNode;
-        screenNode = window()->createImageNode();
-        screenNode->setOwnsTexture(true);
-        pwNode->appendChildNode(screenNode);
-    } else {
-        screenNode = static_cast<QSGImageNode *>(pwNode->childAtIndex(0));
     }
+
+    QSGImageNode *screenNode = pwNode->screenNode(window());
     screenNode->setTexture(texture);
+    screenNode->setOwnsTexture(true);
 
     const auto br = boundingRect().toRect();
     QRect rect({0, 0}, texture->textureSize().scaled(br.size(), Qt::KeepAspectRatio));
@@ -287,6 +284,7 @@ QSGNode *PipeWireSourceItem::updatePaintNode(QSGNode *node, QQuickItem::UpdatePa
         QSGImageNode *cursorNode = pwNode->cursorNode(window());
         if (d->m_cursor.dirty || !cursorNode->texture()) {
             cursorNode->setTexture(window()->createTextureFromImage(d->m_cursor.texture));
+            cursorNode->setOwnsTexture(true);
             d->m_cursor.dirty = false;
         }
         const qreal scale = qreal(rect.width()) / texture->textureSize().width();
@@ -306,6 +304,7 @@ QSGNode *PipeWireSourceItem::updatePaintNode(QSGNode *node, QQuickItem::UpdatePa
             p.drawRect(rect);
         }
         damageNode->setTexture(window()->createTextureFromImage(damageImage));
+        damageNode->setOwnsTexture(true);
         damageNode->setRect(rect);
         Q_ASSERT(damageNode->texture());
     }
@@ -335,8 +334,8 @@ void PipeWireSourceItem::processFrame(const PipeWireFrame &frame)
 
     if (frame.dmabuf) {
         updateTextureDmaBuf(*frame.dmabuf, frame.format);
-    } else if (frame.image) {
-        updateTextureImage(*frame.image);
+    } else if (frame.dataFrame) {
+        updateTextureImage(frame.dataFrame);
     }
 
     if (window() && window()->isVisible()) {
@@ -390,21 +389,21 @@ void PipeWireSourceItem::updateTextureDmaBuf(const DmaBufAttributes &attribs, sp
         return QNativeInterface::QSGOpenGLTexture::fromNative(textureId, window(), size, textureOption);
     };
 
-    setEnabled(true);
+    setReady(true);
 }
 
-void PipeWireSourceItem::updateTextureImage(const QImage &image)
+void PipeWireSourceItem::updateTextureImage(const std::shared_ptr<PipeWireFrameData> &data)
 {
     if (!window()) {
         qCWarning(PIPEWIRE_LOGGING) << "pass";
         return;
     }
 
-    d->m_createNextTexture = [this, image] {
-        return window()->createTextureFromImage(image, QQuickWindow::TextureIsOpaque);
+    d->m_createNextTexture = [this, data] {
+        return window()->createTextureFromImage(data->toImage(), QQuickWindow::TextureIsOpaque);
     };
 
-    setEnabled(true);
+    setReady(true);
 }
 
 void PipeWireSourceItem::componentComplete()
@@ -472,4 +471,17 @@ void PipeWireSourceItem::setAllowDmaBuf(bool allowed)
     }
 }
 
+void PipeWireSourceItem::setReady(bool ready)
+{
+    if (d->m_ready != ready) {
+        d->m_ready = ready;
+        Q_EMIT readyChanged();
+    }
+}
+
+bool PipeWireSourceItem::isReady() const
+{
+    return d->m_ready;
+}
+
 #include "moc_pipewiresourceitem.cpp"
