@@ -11,6 +11,7 @@
 #include "logging.h"
 #include "pipewirecore_p.h"
 #include "pwhelpers.h"
+#include "vaapiutils_p.h"
 
 #include <libdrm/drm_fourcc.h>
 #include <spa/utils/result.h>
@@ -66,6 +67,8 @@ struct PipeWireSourceStreamPrivate
 
     bool m_withDamage = false;
     Fraction maxFramerate;
+
+    PipeWireSourceStream::UsageHint usageHint = PipeWireSourceStream::UsageHint::Render;
 };
 
 static const QVersionNumber pwClientVersion = QVersionNumber::fromString(QString::fromUtf8(pw_get_library_version()));
@@ -92,13 +95,51 @@ uint32_t PipeWireSourceStream::spaVideoFormatToDrmFormat(spa_video_format spa_fo
         return DRM_FORMAT_RGBX8888;
     case SPA_VIDEO_FORMAT_ABGR:
         return DRM_FORMAT_RGBA8888;
+    case SPA_VIDEO_FORMAT_GRAY8:
+        return DRM_FORMAT_R8;
     default:
         qCWarning(PIPEWIRE_LOGGING) << "cannot convert spa format to fourcc" << spa_format;
         return DRM_FORMAT_INVALID;
     }
 }
 
-static QHash<spa_video_format, QList<uint64_t>> queryDmaBufModifiers(EGLDisplay display, const QList<spa_video_format> &formats)
+static QString drmFormatName(uint32_t format)
+{
+    return QString::asprintf("%c%c%c%c %s-endian (0x%08x)",
+                             QLatin1Char(format & 0xff).toLatin1(),
+                             QLatin1Char((format >> 8) & 0xff).toLatin1(),
+                             QLatin1Char((format >> 16) & 0xff).toLatin1(),
+                             QLatin1Char((format >> 24) & 0x7f).toLatin1(),
+                             format & DRM_FORMAT_BIG_ENDIAN ? "big" : "little",
+                             format);
+}
+
+spa_video_format drmFormatToSpaVideoFormat(uint32_t drm_format)
+{
+    switch (drm_format) {
+    case DRM_FORMAT_ABGR8888:
+        return SPA_VIDEO_FORMAT_RGBA;
+    case DRM_FORMAT_XBGR8888:
+        return SPA_VIDEO_FORMAT_RGBx;
+    case DRM_FORMAT_ARGB8888:
+        return SPA_VIDEO_FORMAT_BGRA;
+    case DRM_FORMAT_XRGB8888:
+        return SPA_VIDEO_FORMAT_BGRx;
+    case DRM_FORMAT_BGR888:
+        return SPA_VIDEO_FORMAT_BGR;
+    case DRM_FORMAT_RGB888:
+        return SPA_VIDEO_FORMAT_RGB;
+    case DRM_FORMAT_YUYV:
+        return SPA_VIDEO_FORMAT_YUY2;
+    case DRM_FORMAT_R8:
+        return SPA_VIDEO_FORMAT_GRAY8;
+    default:
+        qCWarning(PIPEWIRE_LOGGING) << "cannot convert drm format to spa" << drmFormatName(drm_format);
+        return SPA_VIDEO_FORMAT_UNKNOWN;
+    }
+}
+
+static QHash<spa_video_format, QList<uint64_t>> queryDmaBufModifiers(EGLDisplay display, const QList<spa_video_format> &formats, PipeWireSourceStream::UsageHint usageHint)
 {
     QHash<spa_video_format, QList<uint64_t>> ret;
     ret.reserve(formats.size());
@@ -126,32 +167,42 @@ static QHash<spa_video_format, QList<uint64_t>> queryDmaBufModifiers(EGLDisplay
         uint32_t drm_format = PipeWireSourceStream::spaVideoFormatToDrmFormat(format);
         if (drm_format == DRM_FORMAT_INVALID) {
             qCDebug(PIPEWIRE_LOGGING) << "Failed to find matching DRM format." << format;
-            break;
+            continue;
         }
 
         if (std::find(drmFormats.begin(), drmFormats.end(), drm_format) == drmFormats.end()) {
-            qCDebug(PIPEWIRE_LOGGING) << "Format " << drm_format << " not supported for modifiers.";
+            qCDebug(PIPEWIRE_LOGGING) << "Format " << drmFormatName(drm_format) << " not supported for modifiers.";
             ret[format] = mods;
-            break;
+            continue;
         }
 
         successFormats = eglQueryDmaBufModifiersEXT(display, drm_format, 0, nullptr, nullptr, &count);
         if (!successFormats) {
             qCWarning(PIPEWIRE_LOGGING) << "Failed to query DMA-BUF modifier count.";
             ret[format] = mods;
-            break;
+            continue;
         }
 
-        QList<uint64_t> modifiers(count);
+        QList<uint64_t> queriedModifiers(count);
         if (count > 0) {
-            if (!eglQueryDmaBufModifiersEXT(display, drm_format, count, modifiers.data(), nullptr, &count)) {
+            if (!eglQueryDmaBufModifiersEXT(display, drm_format, count, queriedModifiers.data(), nullptr, &count)) {
                 qCWarning(PIPEWIRE_LOGGING) << "Failed to query DMA-BUF modifiers.";
             }
         }
 
+        QList<uint64_t> usableModifiers(count + 1);
+        if (usageHint == PipeWireSourceStream::UsageHint::EncodeHardware) {
+            auto vaapi = VaapiUtils::instance();
+            std::copy_if(queriedModifiers.begin(), queriedModifiers.end(), std::back_inserter(usableModifiers), [vaapi, drm_format](uint64_t modifier) {
+                return vaapi->supportsModifier(drm_format, modifier);
+            });
+        } else {
+            usableModifiers = queriedModifiers;
+        }
+
         // Support modifier-less buffers
-        modifiers.push_back(DRM_FORMAT_MOD_INVALID);
-        ret[format] = modifiers;
+        usableModifiers.push_back(DRM_FORMAT_MOD_INVALID);
+        ret[format] = usableModifiers;
     }
     return ret;
 }
@@ -224,6 +275,10 @@ buildFormat(spa_pod_builder *builder, spa_video_format format, const QList<uint6
         auto maxFramerate = SPA_FRACTION(requestedMaxFramerate.numerator, requestedMaxFramerate.denominator);
         spa_pod_builder_add(builder, SPA_FORMAT_VIDEO_framerate, SPA_POD_Fraction(&defFramerate), 0);
         spa_pod_builder_add(builder, SPA_FORMAT_VIDEO_maxFramerate, SPA_POD_CHOICE_RANGE_Fraction(&maxFramerate, &minFramerate, &maxFramerate), 0);
+    } else {
+        auto defFramerate = SPA_FRACTION(0, 1);
+        auto maxFramerate = SPA_FRACTION(1200, 1);
+        spa_pod_builder_add(builder, SPA_FORMAT_VIDEO_framerate, SPA_POD_CHOICE_RANGE_Fraction(&defFramerate, &defFramerate, &maxFramerate), 0);
     }
 
     if (modifiers.size() == 1 && modifiers[0] == DRM_FORMAT_MOD_INVALID) {
@@ -279,7 +334,7 @@ void PipeWireSourceStream::onStreamParamChanged(void *data, uint32_t id, const s
                                               SPA_TYPE_OBJECT_ParamBuffers,
                                               SPA_PARAM_Buffers,
                                               SPA_PARAM_BUFFERS_buffers,
-                                              SPA_POD_CHOICE_RANGE_Int(16, 2, 16),
+                                              SPA_POD_CHOICE_RANGE_Int(3, 2, 16),
                                               SPA_PARAM_BUFFERS_align,
                                               SPA_POD_Int(16),
                                               SPA_PARAM_BUFFERS_dataType,
@@ -322,6 +377,21 @@ static void onProcess(void *data)
     stream->process();
 }
 
+PipeWireFrameData::PipeWireFrameData(spa_video_format format, void *data, QSize size, qint32 stride, PipeWireFrameCleanupFunction *cleanup)
+    : format(format)
+    , data(data)
+    , size(size)
+    , stride(stride)
+    , cleanup(cleanup)
+{
+    cleanup->ref();
+}
+
+PipeWireFrameData::~PipeWireFrameData()
+{
+    PipeWireFrameCleanupFunction::unref(cleanup);
+}
+
 QSize PipeWireSourceStream::size() const
 {
     return QSize(d->videoFormat.size.width, d->videoFormat.size.height);
@@ -346,9 +416,6 @@ PipeWireSourceStream::PipeWireSourceStream(QObject *parent)
     : QObject(parent)
     , d(new PipeWireSourceStreamPrivate)
 {
-    qRegisterMetaType<QList<DmaBufPlane>>();
-    qRegisterMetaType<PipeWireCursor>();
-
     pwStreamEvents.version = PW_VERSION_STREAM_EVENTS;
     pwStreamEvents.process = &onProcess;
     pwStreamEvents.state_changed = &PipeWireSourceStream::onStreamStateChanged;
@@ -389,10 +456,20 @@ uint PipeWireSourceStream::nodeId()
     return d->pwNodeId;
 }
 
+PipeWireSourceStream::UsageHint PipeWireSourceStream::usageHint() const
+{
+    return d->usageHint;
+}
+
+void PipeWireSourceStream::setUsageHint(UsageHint hint)
+{
+    d->usageHint = hint;
+}
+
 QList<const spa_pod *> PipeWireSourceStream::createFormatsParams(spa_pod_builder podBuilder)
 {
     const auto pwServerVersion = d->pwCore->serverVersion();
-    const QList<spa_video_format> formats = {
+    static constexpr auto formats = {
         SPA_VIDEO_FORMAT_RGBx,
         SPA_VIDEO_FORMAT_RGBA,
         SPA_VIDEO_FORMAT_BGRx,
@@ -401,6 +478,7 @@ QList<const spa_pod *> PipeWireSourceStream::createFormatsParams(spa_pod_builder
         SPA_VIDEO_FORMAT_BGR,
         SPA_VIDEO_FORMAT_xBGR,
         SPA_VIDEO_FORMAT_ABGR,
+        SPA_VIDEO_FORMAT_GRAY8,
     };
     QList<const spa_pod *> params;
     params.reserve(formats.size() * 2);
@@ -409,8 +487,13 @@ QList<const spa_pod *> PipeWireSourceStream::createFormatsParams(spa_pod_builder
     d->m_allowDmaBuf = d->m_allowDmaBuf && (pwServerVersion.isNull() || (pwClientVersion >= kDmaBufMinVersion && pwServerVersion >= kDmaBufMinVersion));
     const bool withDontFixate = d->m_allowDmaBuf && (pwServerVersion.isNull() || (pwClientVersion >= kDmaBufModifierMinVersion && pwServerVersion >= kDmaBufModifierMinVersion));
 
+    if (!d->m_allowDmaBuf && d->usageHint == UsageHint::EncodeHardware) {
+        qCWarning(PIPEWIRE_LOGGING) << "DMABUF is unsupported but hardware encoding is requested, which requires DMABUF import. This will not work correctly.";
+    }
+
     if (d->m_availableModifiers.isEmpty()) {
-        d->m_availableModifiers = queryDmaBufModifiers(display, formats);
+        static const auto availableModifiers = queryDmaBufModifiers(display, formats, d->usageHint);
+        d->m_availableModifiers = availableModifiers;
     }
 
     for (auto it = d->m_availableModifiers.constBegin(), itEnd = d->m_availableModifiers.constEnd(); it != itEnd; ++it) {
@@ -501,9 +584,17 @@ void PipeWireSourceStream::handleFrame(struct pw_buffer *buffer)
 
             QImage cursorTexture;
             if (bitmap && bitmap->size.width > 0 && bitmap->size.height > 0) {
-                const uint8_t *bitmap_data = SPA_MEMBER(bitmap, bitmap->offset, uint8_t);
-                cursorTexture =
-                    PWHelpers::SpaBufferToQImage(bitmap_data, bitmap->size.width, bitmap->size.height, bitmap->stride, spa_video_format(bitmap->format));
+                const size_t bufferSize = bitmap->stride * bitmap->size.height * 4;
+                void *bufferData = malloc(bufferSize);
+                memcpy(bufferData, SPA_MEMBER(bitmap, bitmap->offset, uint8_t), bufferSize);
+                cursorTexture = PWHelpers::SpaBufferToQImage(static_cast<const uchar *>(bufferData),
+                                                             bitmap->size.width,
+                                                             bitmap->size.height,
+                                                             bitmap->stride,
+                                                             spa_video_format(bitmap->format),
+                                                             new PipeWireFrameCleanupFunction([bufferData] {
+                                                                 free(bufferData);
+                                                             }));
             }
             frame.cursor = {{cursor->position.x, cursor->position.y}, {cursor->hotspot.x, cursor->hotspot.y}, cursorTexture};
         }
@@ -516,21 +607,21 @@ void PipeWireSourceStream::handleFrame(struct pw_buffer *buffer)
         if (spaBuffer->datas->chunk->size == 0) {
             qCDebug(PIPEWIRE_LOGGING) << "skipping empty memfd buffer";
         } else {
-            uint8_t *map =
-                static_cast<uint8_t *>(mmap(nullptr, spaBuffer->datas->maxsize + spaBuffer->datas->mapoffset, PROT_READ, MAP_PRIVATE, spaBuffer->datas->fd, 0));
+            const uint32_t mapEnd = spaBuffer->datas->maxsize + spaBuffer->datas->mapoffset;
+            uint8_t *map = static_cast<uint8_t *>(mmap(nullptr, mapEnd, PROT_READ, MAP_PRIVATE, spaBuffer->datas->fd, 0));
 
             if (map == MAP_FAILED) {
                 qCWarning(PIPEWIRE_LOGGING) << "Failed to mmap the memory: " << strerror(errno);
                 return;
             }
-            QImage img = PWHelpers::SpaBufferToQImage(map,
-                                                      d->videoFormat.size.width,
-                                                      d->videoFormat.size.height,
-                                                      spaBuffer->datas->chunk->stride,
-                                                      d->videoFormat.format);
-            frame.image = img.copy();
-
-            munmap(map, spaBuffer->datas->maxsize + spaBuffer->datas->mapoffset);
+            auto cleanup = [map, mapEnd]() {
+                munmap(map, mapEnd);
+            };
+            frame.dataFrame = std::make_shared<PipeWireFrameData>(d->videoFormat.format,
+                                                                  map,
+                                                                  QSize(d->videoFormat.size.width, d->videoFormat.size.height),
+                                                                  spaBuffer->datas->chunk->stride,
+                                                                  new PipeWireFrameCleanupFunction(cleanup));
         }
     } else if (spaBuffer->datas->type == SPA_DATA_DmaBuf) {
         DmaBufAttributes attribs;
@@ -555,20 +646,19 @@ void PipeWireSourceStream::handleFrame(struct pw_buffer *buffer)
         if (spaBuffer->datas->chunk->size == 0) {
             qCDebug(PIPEWIRE_LOGGING) << "skipping empty memptr buffer";
         } else {
-            frame.image = PWHelpers::SpaBufferToQImage(static_cast<uint8_t *>(spaBuffer->datas->data),
-                                                       d->videoFormat.size.width,
-                                                       d->videoFormat.size.height,
-                                                       spaBuffer->datas->chunk->stride,
-                                                       d->videoFormat.format);
+            frame.dataFrame = std::make_shared<PipeWireFrameData>(d->videoFormat.format,
+                                                                  spaBuffer->datas->data,
+                                                                  QSize(d->videoFormat.size.width, d->videoFormat.size.height),
+                                                                  spaBuffer->datas->chunk->stride,
+                                                                  nullptr);
         }
     } else {
-        if (spaBuffer->datas->type == SPA_ID_INVALID)
-            qWarning() << "invalid buffer type";
-        else
-            qWarning() << "unsupported buffer type" << spaBuffer->datas->type;
-        QImage errorImage(200, 200, QImage::Format_ARGB32_Premultiplied);
-        errorImage.fill(Qt::red);
-        frame.image = errorImage;
+        if (spaBuffer->datas->type == SPA_ID_INVALID) {
+            qCWarning(PIPEWIRE_LOGGING) << "invalid buffer type";
+        } else {
+            qCWarning(PIPEWIRE_LOGGING) << "unsupported buffer type" << spaBuffer->datas->type;
+        }
+        frame.dataFrame = {};
     }
 
     Q_EMIT frameReceived(frame);
